#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Dec 15 22:23:40 2021

@author: anton
"""

from SLiCAP import *

prj = initProject("Phantom zero Bandwidth Limitation")

i1 = instruction()
i1.setCircuit("trimpBwLimit.cir")
i1.setSimType("numeric")
i1.setSource("I1")
i1.setDetector("V_2")
i1.setLGref("E1")

# Calculate the poles and the zeros of the loop gain:
    
i1.setDataType("pz")
i1.setGainType("loopgain")

PZ = i1.execute()
listPZ(PZ)

# Calculation of the achievable bandwidth

i1.setDataType("laplace")
loopGainLaplace = i1.execute().laplace
gain, coeffsNumer, coeffsDenom = coeffsTransfer(loopGainLaplace)

orderNumer = len(coeffsNumer) - 1
orderDenom = len(coeffsDenom) - 1

LPproduct = np.abs(gain*coeffsNumer[-1]/coeffsDenom[-1])
LPorder = orderDenom-orderNumer
Bw = 1/2/np.pi*LPproduct**(1/LPorder)
print(Bw)

# Here I assume no zeros!
sumOfPoles = 0
for i in range(len(PZ.poles)):
    sumOfPoles += PZ.poles[i]/2/np.pi
    
# Phantom zero compensation for two poles (did not check this)
    
z_phz = - Bw**2/(np.sqrt(2)*Bw+sumOfPoles)

c_phz = -1/(2*np.pi*z_phz*i1.getParValue('R_f'))

print(c_phz)

# plot Root Locus
i1.defPar("C_f", c_phz)
i1.setDataType("pz")

pzLoopGain = i1.execute()

i1.setGainType("servo")
i1.setDataType("poles")
i1.setStepVar("A")
i1.setStepStart("0")
i1.setStepStop("-10M")
i1.setStepMethod("lin")
i1.setStepNum(200)
i1.stepOn()

polesServo = i1.execute()

figRLcomp = plotPZ("RLcomp", "Root locus of compensated amplifier", [pzLoopGain, polesServo], show=True, xmin=-500, xmax=0, xscale="k", ymin=-250, ymax=250, yscale="k")

# Notice, C_f changes the loop gain-poles product and its value needs to be adjusted

# Plot the magnitude and phase characteristics of the transfers of the asymptotic-gain model

i1.setDataType("laplace")
i1.stepOff()
i1.setGainType("asymptotic")
AG = i1.execute()
i1.setGainType("loopgain")
LG = i1.execute()
i1.setGainType("servo")
SG = i1.execute()
i1.setGainType("direct")
DG = i1.execute()
i1.setGainType("gain")
G  = i1.execute()

figMagComp = plotSweep("magComp", "Magnitude of the transfers of the asymptotic gain model", [AG, LG, SG, DG, G], 0.01, 10, 100, sweepScale="M", funcType="mag", show=True)
figPhaComp = plotSweep("phaComp", "Phase of the transfers of the asymptotic gain model", [AG, LG, SG, DG, G], 0.01, 10, 100, sweepScale="M", funcType="phase", show=True)

# Bandwidth limitation with phantom zero
# Limit the bandwidth to 100 kHz

c_lim = 1/(2*np.pi*100e3*i1.getParValue('R_f'))
i1.defPar("C_f", c_lim)

# plot the root locus
i1.setDataType("pz")

pzLoopGainBw = i1.execute()

i1.setGainType("servo")
i1.setDataType("poles")
i1.stepOn()

polesServoBw = i1.execute()

figRLbwLim = plotPZ("RLbwLim", "Root locus of amplifier with reduced bandwidth", [pzLoopGainBw, polesServoBw], show=True, xmin=-500, xmax=0, xscale="k", ymin=-250, ymax=250, yscale="k")

# Notice that the loop gain at 100kHz is not large enough for the pole to reach the zero
# The bandwidth is larger than 100kHz

# Plot the root locus with C_f as root locus variable to see its effect
i1.setStepVar("C_f")
i1.setStepStart(c_phz)
i1.setStepStop(c_lim)
polesServoBwCf = i1.execute()

figRLbwLimCf = plotPZ("RLbwLimCf", "Root locus of amplifier with reduced bandwidth versus $C_f$", polesServoBwCf, show=True, xmin=-500, xmax=0, xscale="k", ymin=-250, ymax=250, yscale="k")

# make less steps and see the result in the magnitude and phase of the gain
i1.setDataType("laplace")
i1.setGainType("gain")
i1.setStepNum(5)
gainBwLimit = i1.execute()

figMagBwLimit = plotSweep("magCompBw", "Magnitude of the gain", gainBwLimit, 0.01, 10, 100, sweepScale="M", funcType="mag", show=True)
figPhaBwLimit= plotSweep("phaCompBw", "Phase of the gain", gainBwLimit, 0.01, 10, 100, sweepScale="M", funcType="phase", show=True)
